;; $Id: keymap-help.el,v 1.6 1993/06/26 03:38:45 nickson Exp $

;; keymap-help:  give help on prefix keymaps.

;;    Copyright (C) 1993 by Ray Nickson (nickson@cs.uq.oz.au)
;;    This program is free software; you can redistribute it and/or modify
;;    it under the terms of the GNU General Public License as published by
;;    the Free Software Foundation; either version 1, or (at your option)
;;    any later version.
;;
;;    This program is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;    GNU General Public License for more details.
;;
;;    You should have received a copy of the GNU General Public License
;;    along with this program; if not, write to the Free Software
;;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;  Bind this to C-H in every keymap you can think of, e.g. in .emacs do:
;; 
;; (require 'keymap-help)
;; (global-set-key "\C-C\C-H" 'keymap-help)
;; (global-set-key "\C-X\C-H" 'keymap-help)
;; (global-set-key "\C-[\C-H" 'keymap-help)  ;was: mark-defun
;; (global-set-key "\C-X4\C-H" 'keymap-help)
;; (global-set-key "\C-X5\C-H" 'keymap-help)
;; (global-set-key "\C-X6\C-H" 'keymap-help)
;; (global-set-key "\C-XX\C-H" 'keymap-help)
;; (global-set-key "\C-Xa\C-H" 'keymap-help)
;; (global-set-key "\C-Xai\C-H" 'keymap-help)
;; (global-set-key "\C-Xn\C-H" 'keymap-help)
;; (global-set-key "\C-Xr\C-H" 'keymap-help)
;; (global-set-key "\C-Xv\C-H" 'keymap-help)
;; (global-set-key "\C-X\C-[\C-H" 'keymap-help)

(provide 'keymap-help)

(defvar keymap-help-last-keys nil)
(defun keymap-help ()
  "Give help on the keymap for the prefix in which this key is bound."
  (interactive)
  (let ((keys (this-command-keys)))
    (setq keys (substring keys 0 (1- (length keys))))
    (if (and (eq last-command 'keymap-help) keymap-help-last-keys)
        (setq keys keymap-help-last-keys))
    (if (and (not keymap-help-last-keys)
             (keymap-help-brief (setq keymap-help-last-keys keys)))
        (setq unread-command-events (string-to-chars keys))
      (setq keymap-help-last-keys nil)
      (set-buffer (get-buffer-create "*Help*"))
      (erase-buffer)
      (keymap-help-full keys (key-binding keys))
      (display-buffer "*Help*"))))

(defvar keymap-help-alist
  '(("\C-X4" .
 "other window: [b]uffer, [f]ile, [d]ired, [r]ead-only, v=scroll, .=find-tag")
    ("\C-X5" .
 "frame: 0=delete, 2=new, [f]ile, [b]uffer, [d]ired, [r]ead-only, .=find-tag")
    ("\C-X6" .
 "two-column: 2, 1, RET=scroll-line, DEL=scroll-down, SPC=scroll-up, [s]plit")
    ("\C-XX" .
 "edebug: [w]here, [b]reak, [u]nbreak, [c]ont, [t]race, [g]o, [d]efun")
    ("\C-Xa" .
 "abbrev: [e]xpand, [g]lobal-add, [l]ocal-add, [i]nverse")
    ("\C-Xai" .
 "add inverse abbrev: [g]lobal, [l]ocal")
    ("\C-Xn" .
 "narrow: [p]age, n=region, [w]iden")
    ("\C-Xv" .
 "vc: v=next, u=revert, [s]napshot, [r]etrieve, [l]og, [c]ancel")
    ("\C-Xr" .
 "register and rectangle")
    )
  "Alist of (KEYSEQ . MESSAGE) pairs for brief keymap help."
)

; this is useful for user changes to keymap-help-alist
(defun add-assoc (alist &rest assocs)
  "Modify ALIST by adding ASSOCS (rest of args).  Replace associations ifkeys
match.  ALIST must be a quoted atom whose value is an alist; it will be
modified by side-effect if possible."
  (while assocs
    (let ((assoc (assoc (car (car assocs)) (eval alist))))
      (if assoc (rplacd assoc (cdr (car assocs)))
        (set alist (cons (car assocs) (eval alist)))))
    (setq assocs (cdr assocs))))

(defun keymap-help-brief (keys)
  "Give brief help on KEYS.  Not yet implemented."
  (let ((help (assoc keys keymap-help-alist)))
    (and help (message (cdr help)))))

(defun keymap-help-full (keys map)
  "Give full help on KEYS bound to MAP."
  (if (symbolp map)
      (setq map (symbol-function map)))
  (if (vectorp (car (cdr map)))
        (keymap-help-vector keys (car (cdr map)))
      (keymap-help-sparse keys (cdr map))))

(defun keymap-help-vector (keys vec)
  "KEYS invoke vector keymap VEC.  Describe it."
  (let ((n 0))
    (while (< n 128)
      (let ((ent (aref vec n)))
        (cond ((null ent)
               nil)
              ((keymapp ent)
               (keymap-help-full (concat keys (char-to-string n)) ent))
              (t
               (insert (key-description (concat keys (char-to-string n))))
               (insert-char ?  (- 20 (current-column)))
               (insert (symbol-name ent) "\n"))))
      (setq n (1+ n)))))

(defun keymap-help-sparse (keys alist)
  "KEYS invoke sparse keymap ALIST.  Describe it."
  (while alist
    (let ((ent (car alist)))
      (if (keymapp (cdr ent))
          (keymap-help-full (concat keys (char-to-string (car ent))) (cdr ent))
        (insert (key-description (concat keys (char-to-string (car ent)))))
        (insert-char ?  (- 20 (current-column)))
        (insert (symbol-name (cdr ent)) "\n")))
    (setq alist (cdr alist))))
                 
(defun string-to-chars (str)
  "Convert STRING to a list of characters."
  (let ((n (length str)) (res nil))
    (while (> n 0)
      (setq res (cons (aref str (1- n)) res)
            n (1- n)))
    res))
      
